/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operate_ManipulatorCommon_h__
#define __operate_ManipulatorCommon_h__

namespace fe
{

/**************************************************************************//**
    @brief Partial Generic ManipulatorI Implemention

	@ingroup operate
*//***************************************************************************/
class FE_DL_EXPORT ManipulatorCommon:
	public OperateCommon,
	virtual public ManipulatorI,
	public Initialize<ManipulatorCommon>
{
	public:
					ManipulatorCommon(void):
						m_hotGrip(-1),
						m_lastGrip(-1),
						m_dragging(FALSE),
						m_anchorScale(1.0),
						m_gripScale(1.0),
						m_dimming(0.0),
						m_foreshadow(0.0),
						m_aura(0.0)											{}
virtual				~ManipulatorCommon(void)								{}

		void		initialize(void);

					//* As ManipulatorI
virtual	void		bindOverlay(sp<DrawI>& a_spDrawOverlay);

virtual	BWORD		gripLabels(void) const	{ return m_gripLabels; }
virtual	void		setGripLabels(BWORD a_gripLabels)
					{	m_gripLabels=a_gripLabels; }

virtual	Real		anchorScale(void) const	{ return m_anchorScale; }
virtual	void		setAnchorScale(Real a_anchorScale)
					{	m_anchorScale=a_anchorScale; }

virtual	Real		gripScale(void) const	{ return m_gripScale; }
virtual	void		setGripScale(Real a_gripScale)
					{	m_gripScale=a_gripScale; }

virtual	Real		dimming(void) const	{ return m_dimming; }
virtual	void		setDimming(Real a_dimming)
					{	m_dimming=a_dimming; }

virtual	Real		foreshadow(void) const	{ return m_foreshadow; }
virtual	void		setForeshadow(Real a_foreshadow)
					{	m_foreshadow=a_foreshadow; }

virtual	Real		aura(void) const	{ return m_aura; }
virtual	void		setAura(Real a_aura)
					{	m_aura=a_aura; }

					//* As HandlerI (window events)
virtual	void		handle(Record& a_rSignal);

					//* As DrawableI
virtual	void		draw(sp<DrawI> a_spDrawI,const Color *color);
virtual	void		draw(sp<DrawI> a_spDrawI,const Color *color) const		{}
virtual void		draw(const SpatialTransform &transform,
							sp<DrawI> a_spDrawI,const Color* color)			{}
virtual void		draw(const SpatialTransform &transform,
							sp<DrawI> a_spDrawI,const Color* color) const	{}

	protected:

	class Grip
	{
		public:
							Grip(void):
								m_hotSpot(0.0,0.0,0.0),
								m_radius(0.0)
							{
								set(m_alignment[0],0.0,0.0,0.0);
								set(m_alignment[1],0.0,0.0,0.0);
								set(m_planeCenter,0.0,0.0,0.0);
								set(m_planeFacing,0.0,0.0,0.0);
							}

							//* screen-space
			SpatialVector	m_hotSpot;
			Real			m_radius;
			String			m_key[2];
			String			m_message;
			String			m_tag;

							//* world-space
			SpatialVector	m_alignment[2];
			SpatialVector	m_planeCenter;
			SpatialVector	m_planeFacing;
	};

					//* reset hotspots and update the draw cache
virtual	void		updateGrips(void)										{}

		std::vector<Grip>	m_gripArray;
		I32					m_hotGrip;
		I32					m_lastGrip;
		BWORD				m_dragging;
		Vector2i			m_lastMouse;

		sp<DrawI>			m_spDrawCache;
		sp<DrawI>			m_spOverlayCache;
		WindowEvent			m_event;
		BWORD				m_gripLabels;
		Real				m_anchorScale;
		Real				m_gripScale;
		Real				m_dimming;
		Real				m_foreshadow;
		Real				m_aura;
};

} // namespace

#endif /* __operate_ManipulatorCommon_h__ */
